VSVersionInfo(
  ffi=FixedFileInfo(
    filevers=(@@COMMAVERSION@@),
    prodvers=(@@COMMAVERSION@@),
    mask=0x3f,
    flags=0x0,
    OS=0x40004,
    fileType=0x1,
    subtype=0x0,
    date=(0, 0)
    ),
  kids=[
    StringFileInfo(
      [
      StringTable(
        u'040904B0',
        [StringStruct(u'CompanyName', u'EUMETSAT'),
        StringStruct(u'FileDescription', u'EUMETSAT Data Access Client'),
        StringStruct(u'FileVersion', u'@@VERSION@@'),
        StringStruct(u'InternalName', u'eumdac'),
        StringStruct(u'LegalCopyright', u'\xa9 EUMETSAT. All rights reserved.'),
        StringStruct(u'OriginalFilename', u'eumdac.exe'),
        StringStruct(u'ProductName', u'EUMETSAT Data Access Client'),
        StringStruct(u'ProductVersion', u'@@VERSION@@')])
      ]), 
    VarFileInfo([VarStruct(u'Translation', [0x0809, 1252])])
  ]
)