"""Module containing package information for setup.py and __init__.py"""

__title__ = "eumdac"
__description__ = "EUMETSAT Data Access Client"
__url__ = "https://gitlab.eumetsat.int/eumetlab/data-services/eumdac"
__version__ = "3.0.0"
__documentation__ = (
    "https://user.eumetsat.int/resources/user-guides/eumetsat-data-access-client-eumdac-guide"
)
__api_documentation__ = "https://usc.tools.eumetsat.int/docs/eumdac/"
__author__ = "EUMETSAT"
__author_email__ = "ops@eumetsat.int"
__license__ = "MIT"
